// Gmail Manager
// By Todd Long <longfocus@gmail.com>
// http://www.longfocus.com/firefox/gmanager/

function Cookies()
{
  this._console = Components.classes['@mozilla.org/consoleservice;1'].getService(Components.interfaces.nsIConsoleService);
  this._observer = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
  this._sessions = new Array();
}

Cookies.prototype = {
  loadSession: function(aHost)
  {
    if (this._sessions[aHost] == null)
    {
      this._sessions[aHost] = this._getCookies(aHost);
      
      this._observer.addObserver(this, "http-on-modify-request", false);
      this._observer.addObserver(this, "http-on-examine-response", false);
    }
  },
  
  restoreSession: function(aHost)
  {
    if (this._sessions[aHost] != null)
    {
      this._putCookies(aHost);
      delete this._sessions[aHost];
      
      this._observer.removeObserver(this, "http-on-modify-request");
      this._observer.removeObserver(this, "http-on-examine-response");
    }
  },
  
  _getCookies: function(aHost)
  {
    var iioServ = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    var cookieMan = Components.classes["@mozilla.org/cookiemanager;1"].getService(Components.interfaces.nsICookieManager);
    var cookieManEnum = cookieMan.enumerator;
    var cookies = new Array();
    
    while (cookieManEnum.hasMoreElements()) {
      // Use nsICookie2 to access isSession and rawHost
      var nextCookie = cookieManEnum.getNext().QueryInterface(Components.interfaces.nsICookie2);
      
      if (nextCookie.host.indexOf(aHost) > -1) {
        var uri = iioServ.newURI("http://" + nextCookie.rawHost, "UTF-8", null);
        var name = nextCookie.name + "=" + nextCookie.value + ";";
        var type = (nextCookie.isDomain ? "Domain=" : "Host=") + nextCookie.host + ";";
        var path = "Path=" + nextCookie.path + ";";
        var expires = "";
        
        if (!nextCookie.isSession)
          expires = "Expires=" + new Date(nextCookie.expires * 1000).toGMTString() + ";";
        
        cookies[name] = new Object();
        cookies[name].cookie = name + type + path + expires;
        cookies[name].uri = uri;
      }
    }
    
    return cookies;
  },
  
  _putCookies: function(aHost)
  {
    var cookieServ = Components.classes["@mozilla.org/cookieService;1"].getService(Components.interfaces.nsICookieService);
    var cookieMan = Components.classes["@mozilla.org/cookiemanager;1"].getService(Components.interfaces.nsICookieManager);
    var cookieManEnum = cookieMan.enumerator;
    
    while (cookieManEnum.hasMoreElements()) {
      var nextCookie = cookieManEnum.getNext().QueryInterface(Components.interfaces.nsICookie2);
      
      if (nextCookie.host.indexOf(aHost) > -1)
       cookieMan.remove(nextCookie.host, nextCookie.name, nextCookie.path, false);
    }
    
    for (var cookieName in this._sessions[aHost]) {
      var cookie = this._sessions[aHost][cookieName].cookie;
      var uri = this._sessions[aHost][cookieName].uri;
      
      cookieServ.setCookieString(uri, null, cookie, null);
    }
  },
  
  observe: function(aSubject, aTopic, aData)
  {
    var httpChannel = aSubject.QueryInterface(Components.interfaces.nsIHttpChannel);
    var uri = httpChannel.URI;
    var host = null;
    
    // Find matching host
    for (var lookupHost in this._sessions)
      if (uri.host.indexOf(lookupHost) > -1)
        host = lookupHost;
    
    var isGood = (uri.originCharset.indexOf("gmanager") == -1 && host != null);
    
    if (aTopic == "http-on-modify-request" && isGood)
    {
      try {
        httpChannel.setRequestHeader("Cookie", "", false);
        
        for (var cookieName in this._sessions[host]) {
          var cookie = this._sessions[host][cookieName].cookie;
          var cookieValue = (cookie.split(";"))[0];
          var cookieHost = this._sessions[host][cookieName].uri.host;
          
          if (cookie.indexOf("Domain") > -1)
            cookieHost = cookie.match(/\Domain=[a-zA-Z.]*;/).slice(7, -1);
          else if (cookie.indexOf("Host") > -1)
            cookieHost = cookie.match(/\Host=[a-zA-Z.]*;/).slice(5, -1);
          
          if (uri.host.indexOf(cookieHost) > -1 && cookieValue.indexOf("GV=") == -1)
            httpChannel.setRequestHeader("Cookie", cookieValue + ";", true);
        }
      } catch(e) {}
    }
    else if (aTopic == "http-on-examine-response" && isGood)
    {
      try {
        var cookies = httpChannel.getResponseHeader("Set-Cookie").split("\n");
        
        for (var i = 0; i < cookies.length; i++)
        {
          var cookieName = (cookies[i].split("="))[0];
          this._sessions[host][cookieName] = new Object();
          this._sessions[host][cookieName].cookie = cookies[i];
          this._sessions[host][cookieName].uri = uri;
        }
      } catch(e) {}
    }
  },
  
  QueryInterface: function(iid)
  {
    if (iid.equals(Components.interfaces.nsICookies) ||
        iid.equals(Components.interfaces.nsISupports)) return this;
    
    Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
    return null;
  }
}

var myModule = {
  firstTime: true,
  
  myCID: Components.ID("{fb560413-70f0-11da-9537-00e08161165f}"),
  myDesc: "Cookie management for accounts",
  myProgID: "@longfocus.com/gmanager/cookies;1",
  myFactory: {
    createInstance: function (outer, iid) {
      if (outer != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;
      
      return (new Cookies()).QueryInterface(iid);
    }
  },

  registerSelf: function (compMgr, fileSpec, location, type)
  {
    if (this.firstTime) {
      this.firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    compMgr.registerFactoryLocation(this.myCID, this.myDesc, this.myProgID, fileSpec, location, type);
  },
 
  getClassObject: function (compMgr, cid, iid)
  {
    if (!cid.equals(this.myCID))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    
    if (!iid.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    
    return this.myFactory;
  },
  
  canUnload: function(compMgr) { return true; }
};

function NSGetModule(compMgr, fileSpec) { return myModule; }