// Gmail Manager
// By Todd Long <longfocus@gmail.com>
// http://www.longfocus.com/firefox/gmanager/

const GMANAGER_BRANCH = "gmanager.";

const GMAIL_ACCOUNT_STATE_CONNECTING = "connecting";
const GMAIL_ACCOUNT_STATE_LOGGED_OUT = "logged-out";
const GMAIL_ACCOUNT_STATE_LOGGED_IN = "logged-in";
const GMAIL_ACCOUNT_STATE_NEW_MAIL = "new-mail";
const GMAIL_ACCOUNT_STATE_NO_PASSWORD = "no-password";
const GMAIL_ACCOUNT_STATE_ERROR_PASSWORD = "error-password";
const GMAIL_ACCOUNT_STATE_ERROR_CAPTCHA = "error-captcha";
const GMAIL_ACCOUNT_STATE_ERROR_NETWORK = "error-network";
const GMAIL_ACCOUNT_STATE_ERROR_SERVER = "error-server";
const GMAIL_ACCOUNT_STATE_ERROR_TIMEOUT = "error-timeout";

const LOGIN_PAGE = "Sign in to Gmail";
const BAD_CREDENTIALS = "Username and password do not match";
const CAPTCHA_REQUIRED = "letters as they are shown in the image";

const CON_PHASE_1 = "https://www.google.com/accounts/ServiceLoginAuth";
const CON_PHASE_2 = "https://mail.google.com/mail/?search=inbox&view=tl&start=0&init=1";

function GmailAccount()
{
  this._observer = Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService);
  this._console = Components.classes['@mozilla.org/consoleservice;1'].getService(Components.interfaces.nsIConsoleService);
  this._branch = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch(GMANAGER_BRANCH);
  this._google = Components.classes["@longfocus.com/gmanager/cookies;1"].createInstance(Components.interfaces.nsICookies);
}

GmailAccount.prototype = {
  _username: null,
  _password: null,
  _remember: false,
  _loggedIn: false,
  _timer: null,
  _interval: 0,
  _connectionPhase: 0,
  _cookieData: null,
  _isChecking: false,
  _status: GMAIL_ACCOUNT_STATE_LOGGED_OUT,
  _inboxUnread: 0,
  _savedDrafts: 0,
  _spamUnread: 0,
  _labels: null,
  _snippets: null,
  _spaceUsed: "0 MB",
  _percentUsed: "0%",
  _totalSpace: "0 MB",
  
  get username() { return this._username; },
  get password() { return this._password; },
  get loggedIn() { return this._loggedIn; },
  get status() { return this._status; },
  get inboxUnread() { return this._inboxUnread; },
  get savedDrafts() { return this._savedDrafts; },
  get spamUnread() { return this._spamUnread; },
  get spaceUsed() { return this._spaceUsed; },
  get percentUsed() { return this._percentUsed; },
  get totalSpace() { return this._totalSpace; },
  
  getNumLabels: function() { return this._labels ? this._labels.length : 0; },
  getLabel: function(aIndex) { return this._labels[aIndex]; },
  
  getNumSnippets: function() { return this._snippets ? this._snippets.length : 0; },
  getSnippetTime: function(aIndex) { return this._snippets[aIndex].time; },
  getSnippetSender: function(aIndex) { return this._snippets[aIndex].sender; },
  getSnippetSubject: function(aIndex) { return this._snippets[aIndex].subject; },
  getSnippetMsg: function(aIndex) { return this._snippets[aIndex].msg; },
  getSnippetNumLabels: function(aIndex) { return this._snippets[aIndex].labels.length; },
  
  set username(aUsername) { this._username = aUsername; },
  set password(aPassword) { this._password = aPassword; },
  set remember(aRemember) { this._remember = aRemember; },
  
  login: function()
  {
    // Check if a password exists for the account
    if (this._password == null || this._password == "") {
      this._setStatus(GMAIL_ACCOUNT_STATE_NO_PASSWORD);
      return;
    }
    else if (this._isChecking)
      return;
    
    // Initializes the login process
    this._loggedIn = false;
    this._connectionPhase = 0;
    this._setChecking(true);
    
    try {
      this._timer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
      this._timer.initWithCallback(this, 30000, this._timer.TYPE_ONE_SHOT);
    } catch(e) {}
    
    var data = "ltmpl=wsad&ltmplcache=2&continue=https%3A%2F%2Fmail.google.com%2Fmail%2F%3F&service=mail&rm=false&ltmpl=wsad&Email=" +
               encodeURIComponent(this._username) + "&Passwd=" +
               encodeURIComponent(this._password) + "&null=Sign+in";
    
    this.serverRequest(CON_PHASE_1, data);
  },
  
  logout: function()
  {
    this._defaults();
    this._setChecking(false);
    this._setStatus(GMAIL_ACCOUNT_STATE_LOGGED_OUT);
    
    if (this._timer) {
      this._timer.cancel();
      this._timer = null;
    }
  },
  
  check: function()
  {
    if (this._loggedIn)
    {
      // Initializes the login process
      this._connectionPhase = 1;
      this._setChecking(true);
      
      // Sends the server request for inbox
      this.serverRequest(CON_PHASE_2, null);
    }
    else
      this.login();
  },
  
  setTimer: function(aInterval)
  {
    if (this._timer) {
      this._timer.cancel();
      this._timer = null;
    }
    
    if (aInterval > 0) {
      try {
        this._interval = aInterval;
        
        this._timer = Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
        this._timer.initWithCallback(this, aInterval * 60000, this._timer.TYPE_ONE_SHOT);
      } catch(e) {}
    }
  },
  
  notify: function(aTimer)
  {
    if (!this._loggedIn && this._isChecking)
    {
      // Sets the specific error
      this._error(GMAIL_ACCOUNT_STATE_ERROR_TIMEOUT);
      
      // Retry login
      this.setTimer(0.3);
    }
    else
      this.check();
  },
  
  resetUnreadCount: function()
  {
    this._inboxUnread = 0;
    this._savedDrafts = 0;
    this._spamUnread = 0;
    
    this.setTimer(this._interval);
  },
  
  _setChecking: function(aChecking)
  {
    if (aChecking)
    {
      // Loads the Google session
      this._google.loadSession("google.com");
      
      // Sets the status to connecting
      //this._setStatus(GMAIL_ACCOUNT_STATE_CONNECTING);
      
      this._observer.addObserver(this, "http-on-modify-request", false);
      this._observer.addObserver(this, "http-on-examine-response", false);
    }
    else
    {
      // Restores the Google session
      this._google.restoreSession("google.com");
      
      try {
        this._observer.removeObserver(this, "http-on-modify-request");
        this._observer.removeObserver(this, "http-on-examine-response");
      } catch(e) {}
    }
    
    this._isChecking = aChecking;
  },
  
  _setStatus: function(aStatus)
  {
    this._status = aStatus;
    this._observer.notifyObservers(null, this._username, aStatus);
  },
  
  _defaults: function()
  {
    // Resets password if necessary
    if (!this._remember)
      this._password = null;
    
    this._loggedIn = false;
    this._timer = null;
    this._interval = 0;
    this._connectionPhase = 0;
    this._cookieData = null;
    this._isChecking = false;
    
    this._inboxUnread = 0;
    this._savedDrafts = 0;
    this._spamUnread = 0;
    this._labels = null;
    this._snippets = null;
    this._spaceUsed = "0 MB";
    this._percentUsed = "0%";
    this._totalSpace = "0 MB";
  },
  
  _error: function(aError)
  {
    this._defaults();
    this._setChecking(false);
    this._setStatus(aError);
  },
  
  serverRequest: function(aURL, aData)
  {
    var ioService = Components.classes["@mozilla.org/network/io-service;1"].createInstance(Components.interfaces.nsIIOService);
    var uri = ioService.newURI(aURL, "gmanager/" + this._username + "/", null);
    var channel = ioService.newChannelFromURI(uri);
    
    if (aData != null)
    {
      var uploadStream = Components.classes["@mozilla.org/io/string-input-stream;1"].createInstance(Components.interfaces.nsIStringInputStream);
      uploadStream.setData(aData, aData.length);
      
      var uploadChannel = channel.QueryInterface(Components.interfaces.nsIUploadChannel);
      uploadChannel.setUploadStream(uploadStream, "application/x-www-form-urlencoded", -1);
      
      channel.QueryInterface(Components.interfaces.nsIHttpChannel).requestMethod = "POST";
    }
    
    channel.asyncOpen(new this.observer(this), null);
  },
  
  observe: function(aSubject, aTopic, aData)
  {
    var httpChannel = aSubject.QueryInterface(Components.interfaces.nsIHttpChannel);
    var isGood = (httpChannel.originalURI.originCharset.indexOf("gmanager/" + this._username + "/") > -1);
    
    if (aTopic == "http-on-modify-request" && isGood)
    {
      var cookie = (this._cookieData != null ? this._cookieData : "");
      
      httpChannel.setRequestHeader("Cookie", cookie, false);
    }
    else if (aTopic == "http-on-examine-response" && isGood)
    {
      try {
        var cookies = httpChannel.getResponseHeader("Set-Cookie");
        
        if (cookies.indexOf("GX=") > -1)
          this._cookieData = cookies.match(/\GX=.*;/);
        else if (cookies.indexOf("SID=") > -1)
          this._cookieData = cookies.match(/\SID=.*;/);
      } catch(e) {}
    }
  },
  
  callback: function(aData, aRequest)
  {
    // Gets the http channel containing our data
    var httpChannel = aRequest.QueryInterface(Components.interfaces.nsIHttpChannel);
    
    var network = true;
    var status = null;
    var statusText = null;
    
    try {
      // Get the http response status
      status = httpChannel.responseStatus;
      statusText = httpChannel.responseStatusText;
    } catch (e) {
      network = false;
    }
    
    //this._console.logStringMessage(this._username);
    //this._console.logStringMessage(aData);
    //this._console.logStringMessage(status + " - " + statusText)
    //this._console.logStringMessage(network);
    
    if (!network) // Network was disconnected
    {
      // Sets the network error
      this._setChecking(false);
      this._setStatus(GMAIL_ACCOUNT_STATE_ERROR_NETWORK);
      
      // Retry login
      this.setTimer(0.3);
      
      return;
    }
    else if (status == null || status != 200) // Status wasn't good
    {
      // Check status
      if (status >= 500)
        this.check(); // Server error, try again
      else
        this._error(GMAIL_ACCOUNT_STATE_ERROR_SERVER);
      
      return;
    }
    else if (aData.indexOf("Sign in to Gmail") > -1)
    {
      // Sets the account error
      if (aData.indexOf("Username and password do not match") > -1)
      {
        // Reset the password
        this._password = null;
        
        // Sets error for incorrect password
        this._error(GMAIL_ACCOUNT_STATE_ERROR_PASSWORD);
      }
      else // Sets error for captcha required
        this._error(GMAIL_ACCOUNT_STATE_ERROR_CAPTCHA);
      
      return;
    }
    
    // Eveything is good so far, check phase
    switch (++this._connectionPhase)
    {
      case 1:
        this.serverRequest(CON_PHASE_2, null);
        break;
      case 2:
        var newMail = false;
        
        // Resets the timer
        this.setTimer(this._interval);
        
        try {
          // Quota ["qu","68 MB","2610 MB","3%","#006633"]
          var myMatches = aData.match(/\["qu",.*\]/);
          var myValues = eval(myMatches[0]);
          
          this._spaceUsed = myValues[1];
          this._totalSpace = myValues[2];
          this._percentUsed = myValues[3];
        } catch(e) {}
        
        try {
          //  method 1 = ["ds",1,0,0,1,0,0,0] inbox, drafts, spam
          //  method 2 = ["ds",[["inbox",0],["drafts",0],["spam",0]]]
          var myMatches = null;
          var isNew = aData.match(/\["ds",\[\[/);
          
          if (isNew)
            myMatches = aData.match(/\["ds",.*\n(.*]\n)+\]/);
          else
            myMatches = aData.match(/\["ds",.*\]/);
          
          var myValues = eval(myMatches[0]);
          var tempInbox = (isNew) ? myValues[1][0][1] : myValues[1];
          
          // Checks for new mail
          newMail = (tempInbox > this._inboxUnread);
          
          this._inboxUnread = tempInbox;
          this._savedDrafts = (isNew) ? myValues[1][1][1] : myValues[4];
          this._spamUnread = (isNew) ? myValues[1][2][1] : myValues[6];
        } catch(e) {}
        
        try {
          // Labels ["ct",[["label1",0],["label2",0]]]
          var myMatches = aData.match(/\["ct",.*\n(.*]\n)+\]/gm);
          var myValues = eval(myMatches[0]);
          
          this._labels = eval(myValues[1]);
        } catch(e) {}
        
        try {
          // New Mail Snippets
          var myMatches = aData.match(/\["t",.*\n(.*]\n)+\]/gm);
          
          this._snippets = new Array();
          
          for (var i = 0; i < myMatches.length; i++) {
            var mySnippets = eval(myMatches[i]);
            
            for (var j = 1; j < mySnippets.length; j++) {
              var mySnippet = mySnippets[j];
              var snippet = new Object();
              
              // Check snippet for new mail
              if (mySnippet[1] == "1") {
                snippet.id = mySnippet[0];
                snippet.time = this._stripHtml(mySnippet[3]);
                snippet.sender = this._stripHtml(mySnippet[4]);
                snippet.subject = this._replaceHtmlCodes(this._stripHtml(mySnippet[6]));
                snippet.msg = this._replaceHtmlCodes(this._stripHtml(mySnippet[7]));
                snippet.labels = mySnippet[8];
                snippet.files = mySnippet[9]; // Separated by commas
                snippet.date = mySnippet[12];
                
                this._snippets.push(snippet);
              }
            }
          }
        } catch(e) {}
        
        if (newMail)
          this._setStatus(GMAIL_ACCOUNT_STATE_NEW_MAIL);
        
        this._loggedIn = true;
        this._setChecking(false);
        this._setStatus(GMAIL_ACCOUNT_STATE_LOGGED_IN);
        
        break;
    }
  },
  
  observer: function(aThis)
  {
    return ({
      _myData: "",
      
      onStartRequest: function(aRequest, aContext) {
        this._myData = "";
      },
      
      onStopRequest: function(aRequest, aContext, aStatus) {
        aThis.callback(this._myData, aRequest);
      },
      
      onDataAvailable: function(aRequest, aContext, aStream, aSourceOffset, aLength) {
        var scriptableInputStream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
        scriptableInputStream.init(aStream);
        this._myData += scriptableInputStream.read(aLength);
      }});
  },
  
  _stripHtml: function(aString)
  {
    return aString.replace(/(<([^>]+)>)/ig, "");
  },
  
  _replaceHtmlCodes: function(aString)
  {
    var htmlCodes = new Array(
      ["&gt;", ">"], ["&lt;", "<"], ["&#39;", "'"], ["&quot;", "\""],
      ["&amp;", "&"], ["&tilde;", "~"], ["&trade;", "?"], ["&copy;", "?"],
      ["&reg;", "?"], ["&hellip;", ""] );
    
    for (var i = 0; i < htmlCodes.length; i++) {
      var re = new RegExp(htmlCodes[i][0], "g");
      aString = aString.replace(re, htmlCodes[i][1]);
    }
    
    return aString;
  },
  
  QueryInterface: function(iid)
  {
    if (iid.equals(Components.interfaces.nsIGmailAccount) ||
        iid.equals(Components.interfaces.nsISupports)) return this;
    
    Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
    return null;
  }
}

var myModule = {
  firstTime: true,
  
  myCID: Components.ID("{FD67A116-3F89-43d6-8BE6-34941E50CCED}"),
  myDesc: "Gmail account details and connection",
  myProgID: "@longfocus.com/gmanager/account/gmail;1",
  myFactory: {
    createInstance: function (outer, iid) {
      if (outer != null)
        throw Components.results.NS_ERROR_NO_AGGREGATION;
      
      return (new GmailAccount()).QueryInterface(iid);
    }
  },

  registerSelf: function (compMgr, fileSpec, location, type)
  {
    if (this.firstTime) {
      this.firstTime = false;
      throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
    }
    
    compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    compMgr.registerFactoryLocation(this.myCID, this.myDesc, this.myProgID, fileSpec, location, type);
  },
 
  getClassObject: function (compMgr, cid, iid)
  {
    if (!cid.equals(this.myCID))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    
    if (!iid.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
    
    return this.myFactory;
  },
  
  canUnload: function(compMgr) { return true; }
};

function NSGetModule(compMgr, fileSpec) { return myModule; }
